"""
Array Utilities
===============

Defines array utilities objects.

References
----------
-   :cite:`Castro2014a` : Castro, S. (2014). Numpy: Fastest way of computing
    diagonal for each row of a 2d array. Retrieved August 22, 2014, from
    http://stackoverflow.com/questions/26511401/\
numpy-fastest-way-of-computing-diagonal-for-each-row-of-a-2d-array/\
26517247#26517247
-   :cite:`Yorke2014a` : Yorke, R. (2014). Python: Change format of np.array or
    allow tolerance in in1d function. Retrieved March 27, 2015, from
    http://stackoverflow.com/a/23521245/931625
"""

from __future__ import annotations

import functools
import numpy as np
import sys
from collections.abc import KeysView, ValuesView
from contextlib import contextmanager
from dataclasses import fields, is_dataclass, replace
from operator import add, mul, pow, sub, truediv

from colour.constants import DEFAULT_FLOAT_DTYPE, DEFAULT_INT_DTYPE, EPSILON
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Callable,
    Dataclass,
    DType,
    DTypeBoolean,
    DTypeFloating,
    DTypeInteger,
    DTypeNumber,
    Floating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Generator,
    Integer,
    IntegerOrArrayLike,
    IntegerOrNDArray,
    Literal,
    NDArray,
    NestedSequence,
    Number,
    NumberOrArrayLike,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from colour.utilities import (
    attest,
    optional,
    suppress_warnings,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "MixinDataclassFields",
    "MixinDataclassIterable",
    "MixinDataclassArray",
    "MixinDataclassArithmetic",
    "as_array",
    "as_int",
    "as_float",
    "as_int_array",
    "as_float_array",
    "as_int_scalar",
    "as_float_scalar",
    "set_default_int_dtype",
    "set_default_float_dtype",
    "get_domain_range_scale",
    "set_domain_range_scale",
    "domain_range_scale",
    "to_domain_1",
    "to_domain_10",
    "to_domain_100",
    "to_domain_degrees",
    "to_domain_int",
    "from_range_1",
    "from_range_10",
    "from_range_100",
    "from_range_degrees",
    "from_range_int",
    "closest_indexes",
    "closest",
    "interval",
    "is_uniform",
    "in_array",
    "tstack",
    "tsplit",
    "row_as_diagonal",
    "orient",
    "centroid",
    "fill_nan",
    "has_only_nan",
    "ndarray_write",
    "zeros",
    "ones",
    "full",
    "index_along_last_axis",
]


class MixinDataclassFields:
    """
    A mixin providing fields introspection for the :class:`dataclass`-like
    class fields.

    Attributes
    ----------
    -   :meth:`~colour.utilities.MixinDataclassFields.fields`
    """

    @property
    def fields(self) -> Tuple:
        """
        Getter property for the fields of the :class:`dataclass`-like class.

        Returns
        -------
        :class:`tuple`
           Tuple of :class:`dataclass`-like class fields.
        """

        return fields(self)


class MixinDataclassIterable(MixinDataclassFields):
    """
    A mixin providing iteration capabilities over the :class:`dataclass`-like
    class fields.

    Attributes
    ----------
    -   :meth:`~colour.utilities.MixinDataclassIterable.keys`
    -   :meth:`~colour.utilities.MixinDataclassIterable.values`
    -   :meth:`~colour.utilities.MixinDataclassIterable.items`

    Methods
    -------
    -   :meth:`~colour.utilities.MixinDataclassIterable.__iter__`

    Notes
    -----
    -   The :class:`colour.utilities.MixinDataclassIterable` class inherits the
        methods from the following class:

        -   :class:`colour.utilities.MixinDataclassFields`
    """

    @property
    def keys(self) -> Tuple:
        """
        Getter property for the :class:`dataclass`-like class keys, i.e. the
        field names.

        Returns
        -------
        :class:`tuple`
           :class:`dataclass`-like class keys.
        """

        return tuple(field for field, _value in self)

    @property
    def values(self) -> Tuple:
        """
        Getter property for the :class:`dataclass`-like class values, i.e. the
        field values.

        Returns
        -------
        :class:`tuple`
           :class:`dataclass`-like class values.
        """

        return tuple(value for _field, value in self)

    @property
    def items(self) -> Tuple:
        """
        Getter property for  the :class:`dataclass`-like class items, i.e. the
        field names and values.

        Returns
        -------
        :class:`tuple`
           :class:`dataclass`-like class items.
        """

        return tuple((field, value) for field, value in self)

    def __iter__(self) -> Generator:
        """
        Return a generator for the :class:`dataclass`-like class fields.

        Yields
        ------
        Generator
           :class:`dataclass`-like class field generator.
        """

        yield from {
            field.name: getattr(self, field.name) for field in self.fields
        }.items()


class MixinDataclassArray(MixinDataclassIterable):
    """
    A mixin providing conversion methods for :class:`dataclass`-like class
    conversion to :class:`numpy.ndarray` class.

    Methods
    -------
    -   :meth:`~colour.utilities.MixinDataclassArray.__array__`

    Notes
    -----
    -   The :class:`colour.utilities.MixinDataclassArray` class inherits the
        methods from the following classes:

        -   :class:`colour.utilities.MixinDataclassIterable`
        -   :class:`colour.utilities.MixinDataclassFields`
    """

    def __array__(self, dtype: Optional[Type[DTypeNumber]] = None) -> NDArray:
        """
        Implement support for :class:`dataclass`-like class conversion to
        :class:`numpy.ndarray` class.

        A field set to *None* will be filled with `np.nan` according to the
        shape of the first field not set with *None*.

        Parameters
        ----------
        dtype
            :class:`numpy.dtype` to use for conversion to `np.ndarray`, default
            to the :class:`numpy.dtype` defined by
            :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

        Returns
        -------
        :class:`numpy.ndarray`
            :class:`dataclass`-like class converted to :class:`numpy.ndarray`.
        """

        dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

        default = None
        for _field, value in self:
            if value is not None:
                default = full(as_float_array(value).shape, np.nan)
                break

        return tstack(
            [value if value is not None else default for value in self.values],
            dtype=dtype,
        )


class MixinDataclassArithmetic(MixinDataclassArray):
    """
    A mixin providing mathematical operations for :class:`dataclass`-like
    class.

    Methods
    -------
    -   :meth:`~colour.utilities.MixinDataclassArray.__iadd__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__add__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__isub__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__sub__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__imul__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__mul__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__idiv__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__div__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__ipow__`
    -   :meth:`~colour.utilities.MixinDataclassArray.__pow__`
    -   :meth:`~colour.utilities.MixinDataclassArray.arithmetical_operation`

    Notes
    -----
    -   The :class:`colour.utilities.MixinDataclassArithmetic` class inherits
        the methods from the following classes:

        -   :class:`colour.utilities.MixinDataclassArray`
        -   :class:`colour.utilities.MixinDataclassIterable`
        -   :class:`colour.utilities.MixinDataclassFields`
    """

    def __add__(self, a: Any) -> Dataclass:
        """
        Implement support for addition.

        Parameters
        ----------
        a
            Variable :math:`a` to add.

        Returns
        -------
        :class:`dataclass`
            Variable added :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "+")

    def __iadd__(self, a: Any) -> Dataclass:
        """
        Implement support for in-place addition.

        Parameters
        ----------
        a
            Variable :math:`a` to add in-place.

        Returns
        -------
        :class:`dataclass`
            In-place variable added :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "+", True)

    def __sub__(self, a: Any) -> Dataclass:
        """
        Implement support for subtraction.

        Parameters
        ----------
        a
            Variable :math:`a` to subtract.

        Returns
        -------
        :class:`dataclass`
            Variable subtracted :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "-")

    def __isub__(self, a: Any) -> Dataclass:
        """
        Implement support for in-place subtraction.

        Parameters
        ----------
        a
            Variable :math:`a` to subtract in-place.

        Returns
        -------
        :class:`dataclass`
            In-place variable subtracted :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "-", True)

    def __mul__(self, a: Any) -> Dataclass:
        """
        Implement support for multiplication.

        Parameters
        ----------
        a
            Variable :math:`a` to multiply by.

        Returns
        -------
        :class:`dataclass`
            Variable multiplied :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "*")

    def __imul__(self, a: Any) -> Dataclass:
        """
        Implement support for in-place multiplication.

        Parameters
        ----------
        a
            Variable :math:`a` to multiply by in-place.

        Returns
        -------
        :class:`dataclass`
            In-place variable multiplied :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "*", True)

    def __div__(self, a: Any) -> Dataclass:
        """
        Implement support for division.

        Parameters
        ----------
        a
            Variable :math:`a` to divide by.

        Returns
        -------
        :class:`dataclass`
            Variable divided :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "/")

    def __idiv__(self, a: Any) -> Dataclass:
        """
        Implement support for in-place division.

        Parameters
        ----------
        a
            Variable :math:`a` to divide by in-place.

        Returns
        -------
        :class:`dataclass`
            In-place variable divided :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "/", True)

    __itruediv__ = __idiv__
    __truediv__ = __div__

    def __pow__(self, a: Any) -> Dataclass:
        """
        Implement support for exponentiation.

        Parameters
        ----------
        a
            Variable :math:`a` to exponentiate by.

        Returns
        -------
        :class:`dataclass`
            Variable exponentiated :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "**")

    def __ipow__(self, a: Any) -> Dataclass:
        """
        Implement support for in-place exponentiation.

        Parameters
        ----------
        a
            Variable :math:`a` to exponentiate by in-place.

        Returns
        -------
        :class:`dataclass`
            In-place variable exponentiated :class:`dataclass`-like class.
        """

        return self.arithmetical_operation(a, "**", True)

    def arithmetical_operation(
        self, a: Any, operation: str, in_place: Boolean = False
    ) -> Dataclass:
        """
        Perform given arithmetical operation with :math:`a` operand on the
        :class:`dataclass`-like class.

        Parameters
        ----------
        a
            Operand.
        operation
            Operation to perform.
        in_place
            Operation happens in place.

        Returns
        -------
        :class:`dataclass`
            :class:`dataclass`-like class with arithmetical operation
            performed.
        """

        callable_operation = {
            "+": add,
            "-": sub,
            "*": mul,
            "/": truediv,
            "**": pow,
        }[operation]

        if is_dataclass(a):
            a = as_float_array(a)

        values = tsplit(callable_operation(as_float_array(self), a))
        field_values = {field: values[i] for i, field in enumerate(self.keys)}
        field_values.update(
            {field: None for field, value in self if value is None}
        )

        dataclass = replace(self, **field_values)

        if in_place:
            for field in self.keys:
                setattr(self, field, getattr(dataclass, field))
            return self
        else:
            return dataclass


def as_array(
    a: Union[NumberOrArrayLike, NestedSequence[NumberOrArrayLike]],
    dtype: Optional[Type[DType]] = None,
) -> NDArray:
    """
    Convert given variable :math:`a` to :class:`numpy.ndarray` using given
    :class:`numpy.dtype`.

    Parameters
    ----------
    a
        Variable to convert.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.ndarray`
        Variable :math:`a` converted to :class:`numpy.ndarray`.

    Examples
    --------
    >>> as_array([1, 2, 3])  # doctest: +ELLIPSIS
    array([1, 2, 3]...)
    >>> as_array([1, 2, 3], dtype=DEFAULT_FLOAT_DTYPE)
    array([ 1.,  2.,  3.])
    """

    # TODO: Remove when https://github.com/numpy/numpy/issues/5718 is
    # addressed.
    if isinstance(a, (KeysView, ValuesView)):
        a = list(a)

    return np.asarray(a, dtype)


def as_int(
    a: NumberOrArrayLike, dtype: Optional[Type[DTypeInteger]] = None
) -> IntegerOrNDArray:
    """
    Attempt to convert given variable :math:`a` to :class:`numpy.integer`
    using given :class:`numpy.dtype`. If variable :math:`a` is not a scalar or
    0-dimensional, it is converted to :class:`numpy.ndarray`.

    Parameters
    ----------
    a
        Variable to convert.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_INT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.integer` or :class:`numpy.ndarray`
        Variable :math:`a` converted to :class:`numpy.integer`.

    Examples
    --------
    >>> as_int(np.array([1]))
    1
    >>> as_int(np.arange(10))  # doctest: +ELLIPSIS
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]...)
    """

    dtype = cast(Type[DTypeInteger], optional(dtype, DEFAULT_INT_DTYPE))

    args = getattr(DTypeInteger, "__args__")
    attest(
        dtype in args,
        f'"dtype" must be one of the following types: {args}',
    )

    # TODO: Reassess implementation when and if
    # https://github.com/numpy/numpy/issues/11956 is addressed.
    return dtype(np.squeeze(a))  # type: ignore[return-value]


def as_float(
    a: NumberOrArrayLike, dtype: Optional[Type[DTypeFloating]] = None
) -> FloatingOrNDArray:
    """
    Attempt to convert given variable :math:`a` to :class:`numpy.floating`
    using given :class:`numpy.dtype`. If variable :math:`a` is not a scalar or
    0-dimensional, it is converted to :class:`numpy.ndarray`.

    Parameters
    ----------
    a
        Variable to convert.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Variable :math:`a` converted to :class:`numpy.floating`.

    Examples
    --------
    >>> as_float(np.array([1]))
    1.0
    >>> as_float(np.arange(10))
    array([ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9.])
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    args = getattr(DTypeFloating, "__args__")
    attest(
        dtype in args,
        f'"dtype" must be one of the following types: {args}',
    )

    return dtype(a)  # type: ignore[arg-type, return-value]


def as_int_array(
    a: NumberOrArrayLike, dtype: Optional[Type[DTypeInteger]] = None
) -> NDArray:
    """
    Convert given variable :math:`a` to :class:`numpy.ndarray` using given
    :class:`numpy.dtype`.

    Parameters
    ----------
    a
        Variable to convert.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_INT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.ndarray`
        Variable :math:`a` converted to :class:`numpy.ndarray`.

    Examples
    --------
    >>> as_int_array([1.0, 2.0, 3.0])  # doctest: +ELLIPSIS
    array([1, 2, 3]...)
    """

    dtype = cast(Type[DTypeInteger], optional(dtype, DEFAULT_INT_DTYPE))

    args = getattr(DTypeInteger, "__args__")
    attest(
        dtype in args,
        f'"dtype" must be one of the following types: {args}',
    )

    return as_array(a, dtype)


def as_float_array(
    a: NumberOrArrayLike, dtype: Optional[Type[DTypeFloating]] = None
) -> NDArray:
    """
    Convert given variable :math:`a` to :class:`numpy.ndarray` using given
    :class:`numpy.dtype`.

    Parameters
    ----------
    a
        Variable to convert.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.ndarray`
        Variable :math:`a` converted to :class:`numpy.ndarray`.

    Examples
    --------
    >>> as_float_array([1, 2, 3])
    array([ 1.,  2.,  3.])
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    attest(
        dtype in np.sctypes["float"],
        f"\"dtype\" must be one of the following types: {np.sctypes['float']}",
    )

    return as_array(a, dtype)


def as_int_scalar(
    a: NumberOrArrayLike, dtype: Optional[Type[DTypeInteger]] = None
) -> Integer:
    """
    Convert given :math:`a` variable to :class:`numpy.integer` using given
    :class:`numpy.dtype`.

    Parameters
    ----------
    a
        Variable to convert.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_INT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.integer`
        :math:`a` variable converted to :class:`numpy.integer`.

    Examples
    --------
    >>> as_int_scalar(np.array(1))
    1
    """

    a = np.squeeze(as_int_array(a, dtype))

    attest(a.ndim == 0, f'"{a}" cannot be converted to "int" scalar!')

    return cast(Integer, as_int(a, dtype))


def as_float_scalar(
    a: NumberOrArrayLike, dtype: Optional[Type[DTypeFloating]] = None
) -> Floating:
    """
    Convert given :math:`a` variable to :class:`numpy.floating` using given
    :class:`numpy.dtype`.

    Parameters
    ----------
    a
        Variable to convert.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.floating`
        :math:`a` variable converted to :class:`numpy.floating`.

    Examples
    --------
    >>> as_float_scalar(np.array(1))
    1.0
    """

    a = np.squeeze(as_float_array(a, dtype))

    attest(a.ndim == 0, f'"{a}" cannot be converted to "float" scalar!')

    return cast(Floating, as_float(a, dtype))


def set_default_int_dtype(
    dtype: Type[DTypeInteger] = DEFAULT_INT_DTYPE,
) -> None:
    """
    Set *Colour* default :class:`numpy.integer` precision by setting
    :attr:`colour.constant.DEFAULT_INT_DTYPE` attribute with given
    :class:`numpy.dtype` wherever the attribute is imported.

    Parameters
    ----------
    dtype
        :class:`numpy.dtype` to set :attr:`colour.constant.DEFAULT_INT_DTYPE`
        with.

    Notes
    -----
    -   It is possible to define the int precision at import time by setting
        the *COLOUR_SCIENCE__DEFAULT_INT_DTYPE* environment variable, for
        example `set COLOUR_SCIENCE__DEFAULT_INT_DTYPE=int32`.

    Warnings
    --------
    This definition is mostly given for consistency purposes with
    :func:`colour.utilities.set_default_float_dtype` definition but contrary to the
    latter, changing *integer* precision will almost certainly completely break
    *Colour*. With great power comes great responsibility.

    Examples
    --------
    >>> as_int_array(np.ones(3)).dtype  # doctest: +SKIP
    dtype('int64')
    >>> set_default_int_dtype(np.int32)
    >>> as_int_array(np.ones(3)).dtype  # doctest: +SKIP
    dtype('int32')
    >>> set_default_int_dtype(np.int64)
    >>> as_int_array(np.ones(3)).dtype  # doctest: +SKIP
    dtype('int64')
    """

    # TODO: Investigate behaviour on Windows.
    with suppress_warnings(colour_usage_warnings=True):
        for module in sys.modules.values():
            if not hasattr(module, "DEFAULT_INT_DTYPE"):
                continue

            setattr(module, "DEFAULT_INT_DTYPE", dtype)


def set_default_float_dtype(
    dtype: Type[DTypeFloating] = DEFAULT_FLOAT_DTYPE,
) -> None:
    """
    Set *Colour* default :class:`numpy.floating` precision by setting
    :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute with given
    :class:`numpy.dtype` wherever the attribute is imported.

    Parameters
    ----------
    dtype
        :class:`numpy.dtype` to set :attr:`colour.constant.DEFAULT_FLOAT_DTYPE`
        with.

    Warnings
    --------
    Changing *float* precision might result in various *Colour* functionality
    breaking entirely: https://github.com/numpy/numpy/issues/6860. With great
    power comes great responsibility.

    Notes
    -----
    -   It is possible to define the *float* precision at import time by
        setting the *COLOUR_SCIENCE__DEFAULT_FLOAT_DTYPE* environment variable,
        for example `set COLOUR_SCIENCE__DEFAULT_FLOAT_DTYPE=float32`.
    -   Some definition returning a single-scalar ndarray might not honour the
        given *float* precision: https://github.com/numpy/numpy/issues/16353

    Examples
    --------
    >>> as_float_array(np.ones(3)).dtype
    dtype('float64')
    >>> set_default_float_dtype(np.float16)
    >>> as_float_array(np.ones(3)).dtype
    dtype('float16')
    >>> set_default_float_dtype(np.float64)
    >>> as_float_array(np.ones(3)).dtype
    dtype('float64')
    """

    with suppress_warnings(colour_usage_warnings=True):
        for module in sys.modules.values():
            if not hasattr(module, "DEFAULT_FLOAT_DTYPE"):
                continue

            setattr(module, "DEFAULT_FLOAT_DTYPE", dtype)


# TODO: Annotate with "Union[Literal['ignore', 'reference', '1', '100'], str]"
# when Python 3.7 is dropped.
_DOMAIN_RANGE_SCALE = "reference"
"""
Global variable storing the current *Colour* domain-range scale.

_DOMAIN_RANGE_SCALE
"""


def get_domain_range_scale() -> Union[
    Literal["ignore", "reference", "1", "100"], str
]:
    """
    Return the current *Colour* domain-range scale. The following scales are
    available:

    -   **'Reference'**, the default *Colour* domain-range scale which varies
        depending on the referenced algorithm, e.g. [0, 1], [0, 10], [0, 100],
        [0, 255], etc...
    -   **'1'**, a domain-range scale normalised to [0, 1], it is important to
        acknowledge that this is a soft normalisation and it is possible to
        use negative out of gamut values or high dynamic range data exceeding
        1.

    Returns
    -------
    :class:`str`
        *Colour* domain-range scale.

    Warnings
    --------
    -   The **'Ignore'** and **'100'** domain-range scales are for internal
        usage only!
    """

    return _DOMAIN_RANGE_SCALE


def set_domain_range_scale(
    scale: Union[
        Literal["ignore", "reference" "Ignore", "Reference", "1", "100"], str
    ] = "reference"
):
    """
    Set the current *Colour* domain-range scale. The following scales are
    available:

    -   **'Reference'**, the default *Colour* domain-range scale which varies
        depending on the referenced algorithm, e.g. [0, 1], [0, 10], [0, 100],
        [0, 255], etc...
    -   **'1'**, a domain-range scale normalised to [0, 1], it is important to
        acknowledge that this is a soft normalisation and it is possible to
        use negative out of gamut values or high dynamic range data exceeding
        1.

    Parameters
    ----------
    scale
        *Colour* domain-range scale to set.

    Warnings
    --------
    -   The **'Ignore'** and **'100'** domain-range scales are for internal
        usage only!
    """

    global _DOMAIN_RANGE_SCALE

    _DOMAIN_RANGE_SCALE = validate_method(
        str(scale),
        ("ignore", "reference", "1", "100"),
        '"{0}" scale is invalid, it must be one of {1}!',
    )


class domain_range_scale:
    """
    Define context manager and decorator temporarily setting *Colour*
    domain-range scale.

    The following scales are available:

    -   **'Reference'**, the default *Colour* domain-range scale which varies
        depending on the referenced algorithm, e.g. [0, 1], [0, 10], [0, 100],
        [0, 255], etc...
    -   **'1'**, a domain-range scale normalised to [0, 1], it is important to
        acknowledge that this is a soft normalisation and it is possible to
        use negative out of gamut values or high dynamic range data exceeding
        1.

    Parameters
    ----------
    scale
        *Colour* domain-range scale to set.

    Warnings
    --------
    -   The **'Ignore'** and **'100'** domain-range scales are for internal
        usage only!

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('1'):
    ...     to_domain_1(1)
    array(1.0)
    >>> with domain_range_scale('Reference'):
    ...     from_range_1(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     to_domain_1(1)
    array(1.0)
    >>> with domain_range_scale('1'):
    ...     from_range_1(1)
    array(1.0)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     to_domain_1(1)
    array(0.01)
    >>> with domain_range_scale('100'):
    ...     from_range_1(1)
    array(100.0)
    """

    def __init__(
        self,
        scale: Union[
            Literal["ignore", "reference" "Ignore", "Reference", "1", "100"],
            str,
        ],
    ):
        self._scale = scale
        self._previous_scale = get_domain_range_scale()

    def __enter__(self) -> domain_range_scale:
        """Set the new domain-range scale upon entering the context manager."""

        set_domain_range_scale(self._scale)

        return self

    def __exit__(self, *args: Any):
        """Set the previous domain-range scale upon exiting the context manager."""

        set_domain_range_scale(self._previous_scale)

    def __call__(self, function: Callable) -> Any:
        """Call the wrapped definition."""

        @functools.wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with self:
                return function(*args, **kwargs)

        return wrapper


def to_domain_1(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 100,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` to domain **'1'**. The behaviour is as
    follows:

    -   If *Colour* domain-range scale is **'Reference'** or **'1'**, the
        definition is almost entirely by-passed and will conveniently convert
        array :math:`a` to :class:`np.ndarray`.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is divided by
        ``scale_factor``, typically 100.

    Parameters
    ----------
    a
        Array :math:`a` to scale to domain **'1'**.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought to domain **'1'**.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled to domain **'1'**.

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     to_domain_1(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     to_domain_1(1)
    array(1.0)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     to_domain_1(1)
    array(0.01)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype).copy()

    if _DOMAIN_RANGE_SCALE == "100":
        a /= as_float_array(scale_factor)

    return a


def to_domain_10(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 10,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` to domain **'10'**, used by
    *Munsell Renotation System*. The behaviour is as follows:

    -   If *Colour* domain-range scale is **'Reference'**, the
        definition is almost entirely by-passed and will conveniently convert
        array :math:`a` to :class:`np.ndarray`.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is
        multiplied by ``scale_factor``, typically 10.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is
        divided by ``scale_factor``, typically 10.

    Parameters
    ----------
    a
        Array :math:`a` to scale to domain **'10'**.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought to domain **'10'**.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled to domain **'10'**.

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     to_domain_10(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     to_domain_10(1)
    array(10.0)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     to_domain_10(1)
    array(0.1)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype).copy()

    if _DOMAIN_RANGE_SCALE == "1":
        a *= as_float_array(scale_factor)

    if _DOMAIN_RANGE_SCALE == "100":
        a /= as_float_array(scale_factor)

    return a


def to_domain_100(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 100,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` to domain **'100'**. The behaviour is as
    follows:

    -   If *Colour* domain-range scale is **'Reference'** or **'100'**
        (currently unsupported private value only used for unit tests), the
        definition is almost entirely by-passed and will conveniently convert
        array :math:`a` to :class:`np.ndarray`.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is
        multiplied by ``scale_factor``, typically 100.

    Parameters
    ----------
    a
        Array :math:`a` to scale to domain **'100'**.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought to domain **'100'**.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled to domain **'100'**.

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     to_domain_100(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     to_domain_100(1)
    array(100.0)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     to_domain_100(1)
    array(1.0)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype).copy()

    if _DOMAIN_RANGE_SCALE == "1":
        a *= as_float_array(scale_factor)

    return a


def to_domain_degrees(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 360,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` to degrees domain. The behaviour is as
    follows:

    -   If *Colour* domain-range scale is **'Reference'**, the
        definition is almost entirely by-passed and will conveniently convert
        array :math:`a` to :class:`np.ndarray`.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is
        multiplied by ``scale_factor``, typically 360.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is
        multiplied by ``scale_factor`` / 100, typically 360 / 100.

    Parameters
    ----------
    a
        Array :math:`a` to scale to degrees domain.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought to degrees domain.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled to degrees domain.

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     to_domain_degrees(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     to_domain_degrees(1)
    array(360.0)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     to_domain_degrees(1)
    array(3.6)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype).copy()

    if _DOMAIN_RANGE_SCALE == "1":
        a *= as_float_array(scale_factor)

    if _DOMAIN_RANGE_SCALE == "100":
        a *= as_float_array(scale_factor) / 100

    return a


def to_domain_int(
    a: ArrayLike,
    bit_depth: IntegerOrArrayLike = 8,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` to int domain. The behaviour is as follows:

    -   If *Colour* domain-range scale is **'Reference'**, the
        definition is almost entirely by-passed and will conveniently convert
        array :math:`a` to :class:`np.ndarray`.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is
        multiplied by :math:`2^{bit\\_depth} - 1`.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is
        multiplied by :math:`2^{bit\\_depth} - 1`.

    Parameters
    ----------
    a
        Array :math:`a` to scale to int domain.
    bit_depth
        Bit depth, usually *integer* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought to int domain.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled to int domain.

    Notes
    -----
    -   To avoid precision issues and rounding, the scaling is performed on
        *float* numbers.

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     to_domain_int(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     to_domain_int(1)
    array(255.0)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     to_domain_int(1)
    array(2.55)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype).copy()

    maximum_code_value = np.power(2, bit_depth) - 1
    if _DOMAIN_RANGE_SCALE == "1":
        a *= maximum_code_value

    if _DOMAIN_RANGE_SCALE == "100":
        a *= maximum_code_value / 100

    return a


def from_range_1(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 100,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` from range **'1'**. The behaviour is as
    follows:

    -   If *Colour* domain-range scale is **'Reference'** or **'1'**, the
        definition is entirely by-passed.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is multiplied
        by ``scale_factor``, typically 100.

    Parameters
    ----------
    a
        Array :math:`a` to scale from range **'1'**.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought from range **'1'**.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled from range **'1'**.

    Warnings
    --------
    The scale conversion of variable :math:`a` happens in-place, i.e. :math:`a`
    will be mutated!

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     from_range_1(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     from_range_1(1)
    array(1.0)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     from_range_1(1)
    array(100.0)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype)

    if _DOMAIN_RANGE_SCALE == "100":
        a *= as_float_array(scale_factor)

    return a


def from_range_10(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 10,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` from range **'10'**, used by
    *Munsell Renotation System*. The behaviour is as follows:

    -   If *Colour* domain-range scale is **'Reference'**, the
        definition is entirely by-passed.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is
        divided by ``scale_factor``, typically 10.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is
        multiplied by ``scale_factor``, typically 10.

    Parameters
    ----------
    a
        Array :math:`a` to scale from range **'10'**.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought from range **'10'**.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled from range **'10'**.

    Warnings
    --------
    The scale conversion of variable :math:`a` happens in-place, i.e. :math:`a`
    will be mutated!

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     from_range_10(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     from_range_10(1)
    array(0.1)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     from_range_10(1)
    array(10.0)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype)

    if _DOMAIN_RANGE_SCALE == "1":
        a /= as_float_array(scale_factor)

    if _DOMAIN_RANGE_SCALE == "100":
        a *= as_float_array(scale_factor)

    return a


def from_range_100(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 100,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` from range **'100'**. The behaviour is as
    follows:

    -   If *Colour* domain-range scale is **'Reference'** or **'100'**
        (currently unsupported private value only used for unit tests), the
        definition is entirely by-passed.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is
        divided by ``scale_factor``, typically 100.

    Parameters
    ----------
    a
        Array :math:`a` to scale from range **'100'**.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought from range **'100'**.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled from range **'100'**.

    Warnings
    --------
    The scale conversion of variable :math:`a` happens in-place, i.e. :math:`a`
    will be mutated!

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     from_range_100(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     from_range_100(1)
    array(0.01)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     from_range_100(1)
    array(1.0)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype)

    if _DOMAIN_RANGE_SCALE == "1":
        a /= as_float_array(scale_factor)

    return a


def from_range_degrees(
    a: ArrayLike,
    scale_factor: FloatingOrArrayLike = 360,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` from degrees range. The behaviour is as
    follows:

    -   If *Colour* domain-range scale is **'Reference'**, the
        definition is entirely by-passed.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is
        divided by ``scale_factor``, typically 360.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is
        divided by ``scale_factor`` / 100, typically 360 / 100.

    Parameters
    ----------
    a
        Array :math:`a` to scale from degrees range.
    scale_factor
        Scale factor, usually *numeric* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought from degrees range.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled from degrees range.

    Warnings
    --------
    The scale conversion of variable :math:`a` happens in-place, i.e. :math:`a`
    will be mutated!

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     from_range_degrees(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     from_range_degrees(1)  # doctest: +ELLIPSIS
    array(0.0027777...)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     from_range_degrees(1)  # doctest: +ELLIPSIS
    array(0.2777777...)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype)

    if _DOMAIN_RANGE_SCALE == "1":
        a /= as_float_array(scale_factor)

    if _DOMAIN_RANGE_SCALE == "100":
        a /= as_float_array(scale_factor) / 100

    return a


def from_range_int(
    a: ArrayLike,
    bit_depth: IntegerOrArrayLike = 8,
    dtype: Optional[Type[DTypeFloating]] = None,
) -> NDArray:
    """
    Scale given array :math:`a` from int range. The behaviour is as follows:

    -   If *Colour* domain-range scale is **'Reference'**, the
        definition is entirely by-passed.
    -   If *Colour* domain-range scale is **'1'**, array :math:`a` is converted
        to :class:`np.ndarray` and divided by :math:`2^{bit\\_depth} - 1`.
    -   If *Colour* domain-range scale is **'100'** (currently unsupported
        private value only used for unit tests), array :math:`a` is converted
        to :class:`np.ndarray` and divided by :math:`2^{bit\\_depth} - 1`.

    Parameters
    ----------
    a
        Array :math:`a` to scale from int range.
    bit_depth
        Bit depth, usually *integer* but can be a :class:`numpy.ndarray` if
        some axis need different scaling to be brought from int range.
    dtype
        Data type used for the conversion to :class:`np.ndarray`.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` scaled from int range.

    Warnings
    --------
    The scale conversion of variable :math:`a` happens in-place, i.e. :math:`a`
    will be mutated!

    Notes
    -----
    -   To avoid precision issues and rounding, the scaling is performed on
        *float* numbers.

    Examples
    --------
    With *Colour* domain-range scale set to **'Reference'**:

    >>> with domain_range_scale('Reference'):
    ...     from_range_int(1)
    array(1.0)

    With *Colour* domain-range scale set to **'1'**:

    >>> with domain_range_scale('1'):
    ...     from_range_int(1)  # doctest: +ELLIPSIS
    array(0.0039215...)

    With *Colour* domain-range scale set to **'100'** (unsupported):

    >>> with domain_range_scale('100'):
    ...     from_range_int(1)  # doctest: +ELLIPSIS
    array(0.3921568...)
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_float_array(a, dtype)

    maximum_code_value = np.power(2, bit_depth) - 1
    if _DOMAIN_RANGE_SCALE == "1":
        a /= maximum_code_value

    if _DOMAIN_RANGE_SCALE == "100":
        a /= maximum_code_value / 100

    return a


def closest_indexes(a: ArrayLike, b: ArrayLike) -> NDArray:
    """
    Return the array :math:`a` closest element indexes to the reference array
    :math:`b` elements.

    Parameters
    ----------
    a
        Array :math:`a` to search for the closest element indexes.
    b
        Reference array :math:`b`.

    Returns
    -------
    :class:`numpy.ndarray`
        Closest array :math:`a` element indexes.

    Examples
    --------
    >>> a = np.array([24.31357115, 63.62396289, 55.71528816,
    ...               62.70988028, 46.84480573, 25.40026416])
    >>> print(closest_indexes(a, 63))
    [3]
    >>> print(closest_indexes(a, [63, 25]))
    [3 5]
    """

    a = np.ravel(a)[:, np.newaxis]
    b = np.ravel(b)[np.newaxis, :]

    return np.abs(a - b).argmin(axis=0)


def closest(a: ArrayLike, b: ArrayLike) -> NDArray:
    """
    Return the array :math:`a` closest elements to the reference array
    :math:`b` elements.

    Parameters
    ----------
    a
        Array :math:`a` to search for the closest element.
    b
        Reference array :math:`b`.

    Returns
    -------
    :class:`numpy.ndarray`
        Closest array :math:`a` elements.

    Examples
    --------
    >>> a = np.array([24.31357115, 63.62396289, 55.71528816,
    ...               62.70988028, 46.84480573, 25.40026416])
    >>> closest(a, 63)
    array([ 62.70988028])
    >>> closest(a, [63, 25])
    array([ 62.70988028,  25.40026416])
    """

    a = np.array(a)

    return a[closest_indexes(a, b)]


def interval(distribution: ArrayLike, unique: Boolean = True) -> NDArray:
    """
    Return the interval size of given distribution.

    Parameters
    ----------
    distribution
        Distribution to retrieve the interval.
    unique
        Whether to return unique intervals if  the distribution is
        non-uniformly spaced or the complete intervals

    Returns
    -------
    :class:`numpy.ndarray`
        Distribution interval.

    Examples
    --------
    Uniformly spaced variable:

    >>> y = np.array([1, 2, 3, 4, 5])
    >>> interval(y)
    array([ 1.])
    >>> interval(y, False)
    array([ 1.,  1.,  1.,  1.])

    Non-uniformly spaced variable:

    >>> y = np.array([1, 2, 3, 4, 8])
    >>> interval(y)
    array([ 1.,  4.])
    >>> interval(y, False)
    array([ 1.,  1.,  1.,  4.])
    """

    distribution = as_float_array(distribution)
    i = np.arange(distribution.size - 1)

    differences = np.abs(distribution[i + 1] - distribution[i])
    if unique:
        return np.unique(differences)
    else:
        return differences


def is_uniform(distribution: ArrayLike) -> Boolean:
    """
    Return whether given distribution is uniform.

    Parameters
    ----------
    distribution
        Distribution to check the uniformity of.

    Returns
    -------
    :class:`bool`
        Whether distribution uniform.

    Examples
    --------
    Uniformly spaced variable:

    >>> a = np.array([1, 2, 3, 4, 5])
    >>> is_uniform(a)
    True

    Non-uniformly spaced variable:

    >>> a = np.array([1, 2, 3.1415, 4, 5])
    >>> is_uniform(a)
    False
    """

    return True if interval(distribution).size == 1 else False


def in_array(
    a: ArrayLike, b: ArrayLike, tolerance: Number = EPSILON
) -> NDArray:
    """
    Return whether each element of the array :math:`a` is also present in the
    array :math:`b` within given tolerance.

    Parameters
    ----------
    a
        Array :math:`a` to test the elements from.
    b
        The array :math:`b` against which to test the elements of array
        :math:`a`.
    tolerance
        Tolerance value.

    Returns
    -------
    :class:`numpy.ndarray`
        A boolean array with array :math:`a` shape describing whether an
        element of array :math:`a` is present in array :math:`b` within given
        tolerance.

    References
    ----------
    :cite:`Yorke2014a`

    Examples
    --------
    >>> a = np.array([0.50, 0.60])
    >>> b = np.linspace(0, 10, 101)
    >>> np.in1d(a, b)
    array([ True, False], dtype=bool)
    >>> in_array(a, b)
    array([ True,  True], dtype=bool)
    """

    a = as_float_array(a)
    b = as_float_array(b)

    d = np.abs(np.ravel(a) - b[..., np.newaxis])

    return np.reshape(np.any(d <= tolerance, axis=0), a.shape)


def tstack(
    a: Union[ArrayLike, NestedSequence[NumberOrArrayLike]],
    dtype: Optional[Union[Type[DTypeBoolean], Type[DTypeNumber]]] = None,
) -> NDArray:
    """
    Stack given array of arrays :math:`a` along the last axis (tail) to
    produce a stacked array.

    It is used to stack an array of arrays produced by the
    :func:`colour.utilities.tsplit` definition.

    Parameters
    ----------
    a
        Array of arrays :math:`a` to stack along the last axis.
    dtype
        :class:`numpy.dtype` to use for initial conversion to
        :class:`numpy.ndarray`, default to the :class:`numpy.dtype` defined by
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.ndarray`
        Stacked array.

    Examples
    --------
    >>> a = 0
    >>> tstack([a, a, a])
    array([ 0.,  0.,  0.])
    >>> a = np.arange(0, 6)
    >>> tstack([a, a, a])
    array([[ 0.,  0.,  0.],
           [ 1.,  1.,  1.],
           [ 2.,  2.,  2.],
           [ 3.,  3.,  3.],
           [ 4.,  4.,  4.],
           [ 5.,  5.,  5.]])
    >>> a = np.reshape(a, (1, 6))
    >>> tstack([a, a, a])
    array([[[ 0.,  0.,  0.],
            [ 1.,  1.,  1.],
            [ 2.,  2.,  2.],
            [ 3.,  3.,  3.],
            [ 4.,  4.,  4.],
            [ 5.,  5.,  5.]]])
    >>> a = np.reshape(a, (1, 1, 6))
    >>> tstack([a, a, a])
    array([[[[ 0.,  0.,  0.],
             [ 1.,  1.,  1.],
             [ 2.,  2.,  2.],
             [ 3.,  3.,  3.],
             [ 4.,  4.,  4.],
             [ 5.,  5.,  5.]]]])
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_array(a, dtype)

    return np.concatenate([x[..., np.newaxis] for x in a], axis=-1)


def tsplit(
    a: Union[ArrayLike, NestedSequence[NumberOrArrayLike]],
    dtype: Optional[Union[Type[DTypeBoolean], Type[DTypeNumber]]] = None,
) -> NDArray:
    """
    Split given stacked array :math:`a` along the last axis (tail) to produce
    an array of arrays.

    It is used to split a stacked array produced by the
    :func:`colour.utilities.tstack` definition.

    Parameters
    ----------
    a
        Stacked array :math:`a` to split.
    dtype
        :class:`numpy.dtype` to use for initial conversion to
        :class:`numpy.ndarray`, default to the :class:`numpy.dtype` defined by
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Returns
    -------
    :class:`numpy.ndarray`
        Array of arrays.

    Examples
    --------
    >>> a = np.array([0, 0, 0])
    >>> tsplit(a)
    array([ 0.,  0.,  0.])
    >>> a = np.array(
    ...     [[0, 0, 0],
    ...      [1, 1, 1],
    ...      [2, 2, 2],
    ...      [3, 3, 3],
    ...      [4, 4, 4],
    ...      [5, 5, 5]]
    ... )
    >>> tsplit(a)
    array([[ 0.,  1.,  2.,  3.,  4.,  5.],
           [ 0.,  1.,  2.,  3.,  4.,  5.],
           [ 0.,  1.,  2.,  3.,  4.,  5.]])
    >>> a = np.array(
    ...     [[[0, 0, 0],
    ...       [1, 1, 1],
    ...       [2, 2, 2],
    ...       [3, 3, 3],
    ...       [4, 4, 4],
    ...       [5, 5, 5]]]
    ... )
    >>> tsplit(a)
    array([[[ 0.,  1.,  2.,  3.,  4.,  5.]],
    <BLANKLINE>
           [[ 0.,  1.,  2.,  3.,  4.,  5.]],
    <BLANKLINE>
           [[ 0.,  1.,  2.,  3.,  4.,  5.]]])
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    a = as_array(a, dtype)

    return np.array([a[..., x] for x in range(a.shape[-1])])


def row_as_diagonal(a: ArrayLike) -> NDArray:
    """
    Return the rows of given array :math:`a` as diagonal matrices.

    Parameters
    ----------
    a
        Array :math:`a` to returns the rows of as diagonal matrices.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` rows as diagonal matrices.

    References
    ----------
    :cite:`Castro2014a`

    Examples
    --------
    >>> a = np.array(
    ...     [[0.25891593, 0.07299478, 0.36586996],
    ...       [0.30851087, 0.37131459, 0.16274825],
    ...       [0.71061831, 0.67718718, 0.09562581],
    ...       [0.71588836, 0.76772047, 0.15476079],
    ...       [0.92985142, 0.22263399, 0.88027331]]
    ... )
    >>> row_as_diagonal(a)
    array([[[ 0.25891593,  0.        ,  0.        ],
            [ 0.        ,  0.07299478,  0.        ],
            [ 0.        ,  0.        ,  0.36586996]],
    <BLANKLINE>
           [[ 0.30851087,  0.        ,  0.        ],
            [ 0.        ,  0.37131459,  0.        ],
            [ 0.        ,  0.        ,  0.16274825]],
    <BLANKLINE>
           [[ 0.71061831,  0.        ,  0.        ],
            [ 0.        ,  0.67718718,  0.        ],
            [ 0.        ,  0.        ,  0.09562581]],
    <BLANKLINE>
           [[ 0.71588836,  0.        ,  0.        ],
            [ 0.        ,  0.76772047,  0.        ],
            [ 0.        ,  0.        ,  0.15476079]],
    <BLANKLINE>
           [[ 0.92985142,  0.        ,  0.        ],
            [ 0.        ,  0.22263399,  0.        ],
            [ 0.        ,  0.        ,  0.88027331]]])
    """

    d = as_array(a)

    d = np.expand_dims(d, -2)

    return np.eye(d.shape[-1]) * d


def orient(
    a: ArrayLike,
    orientation: Union[Literal["Flip", "Flop", "90 CW", "90 CCW", "180"], str],
) -> Union[NDArray, None]:
    """
    Orient given array :math:`a` according to given orientation.

    Parameters
    ----------
    a
        Array :math:`a` to orient.
    orientation
        Orientation to perform.

    Returns
    -------
    :class:`numpy.ndarray`
        Oriented array.

    Examples
    --------
    >>> a = np.tile(np.arange(5), (5, 1))
    >>> a
    array([[0, 1, 2, 3, 4],
           [0, 1, 2, 3, 4],
           [0, 1, 2, 3, 4],
           [0, 1, 2, 3, 4],
           [0, 1, 2, 3, 4]])
    >>> orient(a, '90 CW')
    array([[0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1],
           [2, 2, 2, 2, 2],
           [3, 3, 3, 3, 3],
           [4, 4, 4, 4, 4]])
    >>> orient(a, 'Flip')
    array([[4, 3, 2, 1, 0],
           [4, 3, 2, 1, 0],
           [4, 3, 2, 1, 0],
           [4, 3, 2, 1, 0],
           [4, 3, 2, 1, 0]])
    """

    orientation = validate_method(
        orientation, ["Flip", "Flop", "90 CW", "90 CCW", "180"]
    )

    if orientation == "flip":
        return np.fliplr(a)
    elif orientation == "flop":
        return np.flipud(a)
    elif orientation == "90 cw":
        return np.rot90(a, 3)
    elif orientation == "90 ccw":
        return np.rot90(a)
    elif orientation == "180":
        return np.rot90(a, 2)
    else:  # pragma: no cover
        return None


def centroid(a: ArrayLike) -> NDArray:
    """
    Return the centroid indexes of given array :math:`a`.

    Parameters
    ----------
    a
        Array :math:`a` to returns the centroid indexes of.

    Returns
    -------
    :class:`numpy.ndarray`
        Array :math:`a` centroid indexes.

    Examples
    --------
    >>> a = np.tile(np.arange(0, 5), (5, 1))
    >>> centroid(a)  # doctest: +ELLIPSIS
    array([2, 3]...)
    """

    a = as_float_array(a)

    a_s = np.sum(a)

    ranges = [np.arange(0, a.shape[i]) for i in range(a.ndim)]
    coordinates = np.meshgrid(*ranges)

    a_ci = []
    for axis in coordinates:
        axis = np.transpose(axis)
        # Aligning axis for N-D arrays where N is normalised to
        # range [3, :math:`\\\infty`]
        for i in range(axis.ndim - 2, 0, -1):
            axis = np.rollaxis(axis, i - 1, axis.ndim)

        a_ci.append(np.sum(axis * a) // a_s)

    return np.array(a_ci).astype(DEFAULT_INT_DTYPE)


def fill_nan(
    a: ArrayLike,
    method: Union[Literal["Interpolation", "Constant"], str] = "Interpolation",
    default: Number = 0,
) -> NDArray:
    """
    Fill given array :math:`a` NaN values according to given method.

    Parameters
    ----------
    a
        Array :math:`a` to fill the NaNs of.
    method
        *Interpolation* method linearly interpolates through the NaN values,
        *Constant* method replaces NaN values with ``default``.
    default
        Value to use with the *Constant* method.

    Returns
    -------
    :class:`numpy.ndarray`
        NaNs filled array :math:`a`.

    Examples
    --------
    >>> a = np.array([0.1, 0.2, np.nan, 0.4, 0.5])
    >>> fill_nan(a)
    array([ 0.1,  0.2,  0.3,  0.4,  0.5])
    >>> fill_nan(a, method='Constant')
    array([ 0.1,  0.2,  0. ,  0.4,  0.5])
    """

    a = np.array(a, copy=True)
    method = validate_method(method, ["Interpolation", "Constant"])

    mask = np.isnan(a)

    if method == "interpolation":
        a[mask] = np.interp(
            np.flatnonzero(mask), np.flatnonzero(~mask), a[~mask]
        )
    elif method == "constant":
        a[mask] = default

    return a


def has_only_nan(a: ArrayLike) -> Boolean:
    """
    Return whether given array :math:`a` contains only NaN values.

    Parameters
    ----------
    a
        Array :math:`a` to check whether it contains only NaN values.

    Returns
    -------
    :class:`bool`
        Whether array :math:`a` contains only NaN values.

    Examples
    --------
    >>> has_only_nan(None)
    True
    >>> has_only_nan([None, None])
    True
    >>> has_only_nan([True, None])
    False
    >>> has_only_nan([0.1, np.nan, 0.3])
    False
    """

    a = as_float_array(a)

    return bool(np.all(np.isnan(a)))


@contextmanager
def ndarray_write(a: ArrayLike) -> Generator:
    """
    Define a context manager setting given array :math:`a` writeable to
    operate one and then read-only.

    Parameters
    ----------
    a
        Array :math:`a` to operate on.

    Yields
    ------
    Generator
        Array :math:`a` operated.

    Examples
    --------
    >>> a = np.linspace(0, 1, 10)
    >>> a.setflags(write=False)
    >>> try:
    ...     a += 1
    ... except ValueError:
    ...     pass
    >>> with ndarray_write(a):
    ...     a +=1
    """

    a = as_float_array(a)

    a.setflags(write=True)

    try:
        yield a
    finally:
        a.setflags(write=False)


def zeros(
    shape: Union[Integer, Tuple[int, ...]],
    dtype: Optional[Type[DTypeNumber]] = None,
    order: Literal["C", "F"] = "C",
) -> NDArray:
    """
    Wrap :func:`np.zeros` definition to create an array with the active
    :class:`numpy.dtype` defined by the
    :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Parameters
    ----------
    shape
        Shape of the new array, e.g., ``(2, 3)`` or ``2``.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.
    order
        Whether to store multi-dimensional data in row-major (C-style) or
        column-major (Fortran-style) order in memory.

    Returns
    -------
    :class:`numpy.ndarray`
        Array of given shape and :class:`numpy.dtype`, filled with zeros.

    Examples
    --------
    >>> zeros(3)
    array([ 0.,  0.,  0.])
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    return np.zeros(shape, dtype, order)


def ones(
    shape: Union[Integer, Tuple[int, ...]],
    dtype: Optional[Type[DTypeNumber]] = None,
    order: Literal["C", "F"] = "C",
) -> NDArray:
    """
    Wrap :func:`np.ones` definition to create an array with the active
    :class:`numpy.dtype` defined by the
    :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Parameters
    ----------
    shape
        Shape of the new array, e.g., ``(2, 3)`` or ``2``.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.
    order
        Whether to store multi-dimensional data in row-major (C-style) or
        column-major (Fortran-style) order in memory.

    Returns
    -------
    :class:`numpy.ndarray`
        Array of given shape and type, filled with ones.

    Examples
    --------
    >>> ones(3)
    array([ 1.,  1.,  1.])
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    return np.ones(shape, dtype, order)


def full(
    shape: Union[Integer, Tuple[int, ...]],
    fill_value: Number,
    dtype: Optional[Type[DTypeNumber]] = None,
    order: Literal["C", "F"] = "C",
) -> NDArray:
    """
    Wrap :func:`np.full` definition to create an array with the active type
    defined by the:attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.

    Parameters
    ----------
    shape
        Shape of the new array, e.g., ``(2, 3)`` or ``2``.
    fill_value
        Fill value.
    dtype
        :class:`numpy.dtype` to use for conversion, default to the
        :class:`numpy.dtype` defined by the
        :attr:`colour.constant.DEFAULT_FLOAT_DTYPE` attribute.
    order
        Whether to store multi-dimensional data in row-major (C-style) or
        column-major (Fortran-style) order in memory.

    Returns
    -------
    :class:`numpy.ndarray`
        Array of given shape and :class:`numpy.dtype`, filled with given value.

    Examples
    --------
    >>> ones(3)
    array([ 1.,  1.,  1.])
    """

    dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

    return np.full(shape, fill_value, dtype, order)


def index_along_last_axis(a: ArrayLike, indexes: ArrayLike) -> NDArray:
    """
    Reduce the dimension of array :math:`a` by one, by using an array of
    indexes to pick elements off the last axis.

    Parameters
    ----------
    a
        Array :math:`a` to be indexed.
    indexes
        *Integer* array with the same shape as `a` but with one dimension
        fewer, containing indices to the last dimension of `a`. All elements
        must be numbers between `0` and `m` - 1.

    Returns
    -------
    :class:`numpy.ndarray`
        Indexed array :math:`a`.

    Raises
    ------
    :class:`ValueError`
        If the array :math:`a` and ``indexes`` have incompatible shapes.
    :class:`IndexError`
        If ``indexes`` has elements outside of the allowed range of 0 to
        `m` - 1 or if it's not an *integer* array.

    Examples
    --------
    >>> a = np.array(
    ...     [[[0.3, 0.5, 6.9],
    ...       [3.3, 4.4, 1.6],
    ...       [4.4, 7.5, 2.3],
    ...       [2.3, 1.6, 7.4]],
    ...      [[2. , 5.9, 2.8],
    ...       [6.2, 4.9, 8.6],
    ...       [3.7, 9.7, 7.3],
    ...       [6.3, 4.3, 3.2]],
    ...      [[0.8, 1.9, 0.7],
    ...       [5.6, 4. , 1.7],
    ...       [6.7, 8.2, 1.7],
    ...       [1.2, 7.1, 1.4]],
    ...      [[4. , 4.8, 8.9],
    ...       [4. , 0.3, 6.9],
    ...       [3.5, 7.1, 4.5],
    ...       [1.4, 1.9, 1.6]]]
    ... )
    >>> indexes = np.array(
    ...     [[2, 0, 1, 1],
    ...      [2, 1, 1, 0],
    ...      [0, 0, 1, 2],
    ...      [0, 0, 1, 2]]
    ... )
    >>> index_along_last_axis(a, indexes)
    array([[ 6.9,  3.3,  7.5,  1.6],
           [ 2.8,  4.9,  9.7,  6.3],
           [ 0.8,  5.6,  8.2,  1.4],
           [ 4. ,  4. ,  7.1,  1.6]])

    This function can be used to compute the result of :func:`np.min` along
    the last axis given the corresponding :func:`np.argmin` indexes.

    >>> indexes = np.argmin(a, axis=-1)
    >>> np.array_equal(
    ...     index_along_last_axis(a, indexes),
    ...     np.min(a, axis=-1)
    ... )
    True

    In particular, this can be used to manipulate the indexes given by
    functions like :func:`np.min` before indexing the array. For example, to
    get elements directly following the smallest elements:

    >>> index_along_last_axis(a, (indexes + 1) % 3)
    array([[ 0.5,  3.3,  4.4,  7.4],
           [ 5.9,  8.6,  9.7,  6.3],
           [ 0.8,  5.6,  6.7,  7.1],
           [ 4.8,  6.9,  7.1,  1.9]])
    """

    a = np.array(a)
    indexes = np.array(indexes)

    if a.shape[:-1] != indexes.shape:
        raise ValueError(
            f"Array and indexes have incompatible shapes: {a.shape} and {indexes.shape}"
        )

    return np.take_along_axis(a, indexes[..., np.newaxis], axis=-1).squeeze(
        axis=-1
    )



"""
Hdr-IPT Colourspace
===================

Defines the *hdr-IPT* colourspace transformations:

-   :attr:`colour.HDR_IPT_METHODS`: Supported *hdr-IPT* colourspace computation
    methods.
-   :func:`colour.XYZ_to_hdr_IPT`
-   :func:`colour.hdr_IPT_to_XYZ`

References
----------
-   :cite:`Fairchild2010` : Fairchild, M. D., & Wyble, D. R. (2010). hdr-CIELAB
    and hdr-IPT: Simple Models for Describing the Color of High-Dynamic-Range
    and Wide-Color-Gamut Images. Proc. of Color and Imaging Conference,
    322-326. ISBN:978-1-62993-215-6
-   :cite:`Fairchild2011` : Fairchild, M. D., & Chen, P. (2011). Brightness,
    lightness, and specifying color in high-dynamic-range scenes and images. In
    S. P. Farnand & F. Gaykema (Eds.), Proc. SPIE 7867, Image Quality and
    System Performance VIII (p. 78670O). doi:10.1117/12.872075
"""

from __future__ import annotations

import numpy as np

from colour.algebra import vector_dot
from colour.colorimetry import (
    lightness_Fairchild2010,
    lightness_Fairchild2011,
    luminance_Fairchild2010,
    luminance_Fairchild2011,
)
from colour.hints import (
    ArrayLike,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Literal,
    NDArray,
    Tuple,
    Union,
)
from colour.models.ipt import (
    MATRIX_IPT_XYZ_TO_LMS,
    MATRIX_IPT_LMS_TO_XYZ,
    MATRIX_IPT_LMS_P_TO_IPT,
    MATRIX_IPT_IPT_TO_LMS_P,
)
from colour.utilities import (
    as_float_array,
    domain_range_scale,
    from_range_1,
    from_range_100,
    to_domain_1,
    to_domain_100,
    validate_method,
)
from colour.utilities.documentation import (
    DocstringTuple,
    is_documentation_building,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "HDR_IPT_METHODS",
    "exponent_hdr_IPT",
    "XYZ_to_hdr_IPT",
    "hdr_IPT_to_XYZ",
]

HDR_IPT_METHODS: Tuple = ("Fairchild 2010", "Fairchild 2011")
if is_documentation_building():  # pragma: no cover
    HDR_IPT_METHODS = DocstringTuple(HDR_IPT_METHODS)
    HDR_IPT_METHODS.__doc__ = """
Supported *hdr-IPT* colourspace computation methods.

References
----------
:cite:`Fairchild2010`, :cite:`Fairchild2011`
"""


def exponent_hdr_IPT(
    Y_s: FloatingOrArrayLike,
    Y_abs: FloatingOrArrayLike,
    method: Union[
        Literal["Fairchild 2011", "Fairchild 2010"], str
    ] = "Fairchild 2011",
) -> FloatingOrNDArray:
    """
    Compute *hdr-IPT* colourspace *Lightness* :math:`\\epsilon` exponent using
    *Fairchild and Wyble (2010)* or *Fairchild and Chen (2011)* method.

    Parameters
    ----------
    Y_s
        Relative luminance :math:`Y_s` of the surround.
    Y_abs
        Absolute luminance :math:`Y_{abs}` of the scene diffuse white in
        :math:`cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *hdr-IPT* colourspace *Lightness* :math:`\\epsilon` exponent.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y_s``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    Examples
    --------
    >>> exponent_hdr_IPT(0.2, 100)  # doctest: +ELLIPSIS
    0.4820209...
    >>> exponent_hdr_IPT(0.2, 100, method='Fairchild 2010')
    ... # doctest: +ELLIPSIS
    1.6891383...
    """

    Y_s = to_domain_1(Y_s)
    Y_abs = as_float_array(Y_abs)
    method = validate_method(method, HDR_IPT_METHODS)

    if method == "fairchild 2010":
        epsilon = 1.38
    else:
        epsilon = 0.59

    lf = np.log(318) / np.log(Y_abs)
    sf = 1.25 - 0.25 * (Y_s / 0.184)
    if method == "fairchild 2010":
        epsilon *= sf * lf
    else:
        epsilon /= sf * lf

    return epsilon


def XYZ_to_hdr_IPT(
    XYZ: ArrayLike,
    Y_s: FloatingOrArrayLike = 0.2,
    Y_abs: FloatingOrArrayLike = 100,
    method: Union[
        Literal["Fairchild 2011", "Fairchild 2010"], str
    ] = "Fairchild 2011",
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *hdr-IPT* colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    Y_s
        Relative luminance :math:`Y_s` of the surround.
    Y_abs
        Absolute luminance :math:`Y_{abs}` of the scene diffuse white in
        :math:`cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.ndarray`
        *hdr-IPT* colourspace array.

    Notes
    -----
    +-------------+-------------------------+---------------------+
    | **Domain**  | **Scale - Reference**   | **Scale - 1**       |
    +=============+=========================+=====================+
    | ``XYZ``     | [0, 1]                  | [0, 1]              |
    +-------------+-------------------------+---------------------+
    | ``Y_s``     | [0, 1]                  | [0, 1]              |
    +-------------+-------------------------+---------------------+

    +-------------+-------------------------+---------------------+
    | **Range**   | **Scale - Reference**   | **Scale - 1**       |
    +=============+=========================+=====================+
    | ``IPT_hdr`` | ``I_hdr`` : [0, 100]    | ``I_hdr`` : [0, 1]  |
    |             |                         |                     |
    |             | ``P_hdr`` : [-100, 100] | ``P_hdr`` : [-1, 1] |
    |             |                         |                     |
    |             | ``T_hdr`` : [-100, 100] | ``T_hdr`` : [-1, 1] |
    +-------------+-------------------------+---------------------+

    -   Input *CIE XYZ* tristimulus values must be adapted to
        *CIE Standard Illuminant D Series* *D65*.

    References
    ----------
    :cite:`Fairchild2010`, :cite:`Fairchild2011`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_hdr_IPT(XYZ)  # doctest: +ELLIPSIS
    array([ 48.3937634...,  42.4499020...,  22.0195403...])
    >>> XYZ_to_hdr_IPT(XYZ, method='Fairchild 2010')  # doctest: +ELLIPSIS
    array([ 30.0287314...,  83.9384506...,  34.9028738...])
    """

    XYZ = to_domain_1(XYZ)
    method = validate_method(method, HDR_IPT_METHODS)

    if method == "fairchild 2010":
        lightness_callable = lightness_Fairchild2010
    else:
        lightness_callable = lightness_Fairchild2011

    e = as_float_array(exponent_hdr_IPT(Y_s, Y_abs, method))[..., np.newaxis]

    LMS = vector_dot(MATRIX_IPT_XYZ_TO_LMS, XYZ)

    # Domain and range scaling has already been handled.
    with domain_range_scale("ignore"):
        LMS_prime = np.sign(LMS) * np.abs(lightness_callable(LMS, e))

    IPT_hdr = vector_dot(MATRIX_IPT_LMS_P_TO_IPT, LMS_prime)

    return from_range_100(IPT_hdr)


def hdr_IPT_to_XYZ(
    IPT_hdr: ArrayLike,
    Y_s: FloatingOrArrayLike = 0.2,
    Y_abs: FloatingOrArrayLike = 100,
    method: Union[
        Literal["Fairchild 2011", "Fairchild 2010"], str
    ] = "Fairchild 2011",
) -> NDArray:
    """
    Convert from *hdr-IPT* colourspace to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    IPT_hdr
        *hdr-IPT* colourspace array.
    Y_s
        Relative luminance :math:`Y_s` of the surround.
    Y_abs
        Absolute luminance :math:`Y_{abs}` of the scene diffuse white in
        :math:`cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +-------------+-------------------------+---------------------+
    | **Domain**  | **Scale - Reference**   | **Scale - 1**       |
    +=============+=========================+=====================+
    | ``IPT_hdr`` | ``I_hdr`` : [0, 100]    | ``I_hdr`` : [0, 1]  |
    |             |                         |                     |
    |             | ``P_hdr`` : [-100, 100] | ``P_hdr`` : [-1, 1] |
    |             |                         |                     |
    |             | ``T_hdr`` : [-100, 100] | ``T_hdr`` : [-1, 1] |
    +-------------+-------------------------+---------------------+
    | ``Y_s``     | [0, 1]                  | [0, 1]              |
    +-------------+-------------------------+---------------------+

    +-------------+-------------------------+---------------------+
    | **Range**   | **Scale - Reference**   | **Scale - 1**       |
    +=============+=========================+=====================+
    | ``XYZ``     | [0, 1]                  | [0, 1]              |
    +-------------+-------------------------+---------------------+

    References
    ----------
    :cite:`Fairchild2010`, :cite:`Fairchild2011`

    Examples
    --------
    >>> IPT_hdr = np.array([48.39376346, 42.44990202, 22.01954033])
    >>> hdr_IPT_to_XYZ(IPT_hdr)  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    >>> IPT_hdr = np.array([30.02873147, 83.93845061, 34.90287382])
    >>> hdr_IPT_to_XYZ(IPT_hdr, method='Fairchild 2010')
    ... # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    """

    IPT_hdr = to_domain_100(IPT_hdr)
    method = validate_method(method, HDR_IPT_METHODS)

    if method == "fairchild 2010":
        luminance_callable = luminance_Fairchild2010
    else:
        luminance_callable = luminance_Fairchild2011

    e = as_float_array(exponent_hdr_IPT(Y_s, Y_abs, method))[..., np.newaxis]

    LMS = vector_dot(MATRIX_IPT_IPT_TO_LMS_P, IPT_hdr)

    # Domain and range scaling has already be handled.
    with domain_range_scale("ignore"):
        LMS_prime = np.sign(LMS) * np.abs(luminance_callable(LMS, e))

    XYZ = vector_dot(MATRIX_IPT_LMS_TO_XYZ, LMS_prime)

    return from_range_1(XYZ)



"""
Defines the unit tests for the
:mod:`colour.models.rgb.transfer_functions.exponent` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    exponent_function_basic,
    exponent_function_monitor_curve,
)
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestExponentFunctionBasic",
    "TestExponentFunctionMonitorCurve",
]


class TestExponentFunctionBasic(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_basic` definition unit tests methods.
    """

    def test_exponent_function_basic(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_basic` definition.
        """

        a = 0.18
        a_p = 0.0229932049927
        self.assertAlmostEqual(exponent_function_basic(a, 2.2), a_p, places=7)

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicMirrorFwd"), a_p, places=7
        )

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicPassThruFwd"), a_p, places=7
        )

        a = 0.0229932049927
        a_p = 0.18
        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicRev"), a_p, places=7
        )

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicMirrorRev"), a_p, places=7
        )

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicPassThruRev"), a_p, places=7
        )

        a = -0.18
        self.assertAlmostEqual(exponent_function_basic(a, 2.2), 0.0, places=7)

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicMirrorFwd"),
            -0.0229932049927,
            places=7,
        )

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicPassThruFwd"),
            -0.18,
            places=7,
        )

        a = -0.0229932049927
        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicRev"), 0.0, places=7
        )

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicMirrorRev"), -0.18, places=7
        )

        self.assertAlmostEqual(
            exponent_function_basic(a, 2.2, "basicPassThruRev"),
            -0.0229932049927,
            places=7,
        )

    def test_n_dimensional_exponent_function_basic(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_basic` definition n-dimensional arrays support.
        """

        a = 0.18
        a_p = 0.0229932049927

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicMirrorFwd"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicPassThruFwd"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicMirrorFwd"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicPassThruFwd"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicMirrorFwd"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicPassThruFwd"), a_p, decimal=7
        )

        a = 0.0229932049927
        a_p = 0.18

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicRev"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicMirrorRev"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicPassThruRev"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicRev"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicMirrorRev"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicPassThruRev"), a_p, decimal=7
        )

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicRev"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicMirrorRev"), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_basic(a, 2.2, "basicPassThruRev"), a_p, decimal=7
        )

    @ignore_numpy_errors
    def test_nan_exponent_function_basic(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_basic` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]

        for case in cases:
            exponent_function_basic(case, case)


class TestExponentFunctionMonitorCurve(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_monitor_curve` definition unit tests methods.
    """

    def test_exponent_function_monitor_curve(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_monitor_curve` definition.
        """

        a = 0.18
        a_p = 0.0232240466001
        self.assertAlmostEqual(
            exponent_function_monitor_curve(a, 2.2, 0.001), a_p, places=7
        )

        self.assertAlmostEqual(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorFwd"
            ),
            a_p,
            places=7,
        )

        a = 0.0232240466001
        a_p = 0.18
        self.assertAlmostEqual(
            exponent_function_monitor_curve(a, 2.2, 0.001, "monCurveRev"),
            a_p,
            places=7,
        )

        self.assertAlmostEqual(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorRev"
            ),
            a_p,
            places=7,
        )

        a = -0.18
        self.assertAlmostEqual(
            exponent_function_monitor_curve(a, 2.2, 0.001),
            -0.000205413951,
            places=7,
        )

        self.assertAlmostEqual(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorFwd"
            ),
            -0.0232240466001,
            places=7,
        )

        a = -0.000205413951
        self.assertAlmostEqual(
            exponent_function_monitor_curve(a, 2.2, 0.001, "monCurveRev"),
            -0.18,
            places=7,
        )

        self.assertAlmostEqual(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorRev"
            ),
            -0.0201036111565,
            places=7,
        )

    def test_n_dimensional_exponent_function_monitor_curve(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_monitor_curve` definition n-dimensional arrays support.
        """

        a = 0.18
        a_p = 0.0232240466001

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(a, 2.2, 0.001), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorFwd"
            ),
            a_p,
            decimal=7,
        )

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(a, 2.2, 0.001), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorFwd"
            ),
            a_p,
            decimal=7,
        )

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(a, 2.2, 0.001), a_p, decimal=7
        )
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorFwd"
            ),
            a_p,
            decimal=7,
        )

        a = 0.0232240466001
        a_p = 0.18

        a = np.tile(a, 6)
        a_p = np.tile(a_p, 6)
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(a, 2.2, 0.001, "monCurveRev"),
            a_p,
            decimal=7,
        )
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorRev"
            ),
            a_p,
            decimal=7,
        )

        a = np.reshape(a, (2, 3))
        a_p = np.reshape(a_p, (2, 3))
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(a, 2.2, 0.001, "monCurveRev"),
            a_p,
            decimal=7,
        )
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorRev"
            ),
            a_p,
            decimal=7,
        )

        a = np.reshape(a, (2, 3, 1))
        a_p = np.reshape(a_p, (2, 3, 1))
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(a, 2.2, 0.001, "monCurveRev"),
            a_p,
            decimal=7,
        )
        np.testing.assert_almost_equal(
            exponent_function_monitor_curve(
                a, 2.2, 0.001, "monCurveMirrorRev"
            ),
            a_p,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_exponent_function_monitor_curve(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.exponent.\
exponent_function_monitor_curve` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]

        for case in cases:
            exponent_function_monitor_curve(case, case, case)


if __name__ == "__main__":
    unittest.main()



from .cie import CONSTANT_K_M, CONSTANT_KP_M
from .codata import (
    CONSTANT_AVOGADRO,
    CONSTANT_BOLTZMANN,
    CONSTANT_LIGHT_SPEED,
    CONSTANT_PLANCK,
)
from .common import (
    FLOATING_POINT_NUMBER_PATTERN,
    INTEGER_THRESHOLD,
    EPSILON,
    DEFAULT_INT_DTYPE,
    DEFAULT_FLOAT_DTYPE,
)

__all__ = [
    "CONSTANT_K_M",
    "CONSTANT_KP_M",
]
__all__ += [
    "CONSTANT_AVOGADRO",
    "CONSTANT_BOLTZMANN",
    "CONSTANT_LIGHT_SPEED",
    "CONSTANT_PLANCK",
]
__all__ += [
    "FLOATING_POINT_NUMBER_PATTERN",
    "INTEGER_THRESHOLD",
    "EPSILON",
    "DEFAULT_INT_DTYPE",
    "DEFAULT_FLOAT_DTYPE",
]



"""
Chromatic Adaptation Transforms
===============================

Defines various chromatic adaptation transforms (CAT):

-   :attr:`colour.adaptation.CAT_XYZ_SCALING`: *XYZ Scaling* chromatic
    adaptation transform.
-   :attr:`colour.adaptation.CAT_VON_KRIES`: *Von Kries* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_BRADFORD`: *Bradford* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_SHARP`: *Sharp* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_FAIRCHILD`: *Fairchild* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_CMCCAT97`: *CMCCAT97* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_CMCCAT2000`: *CMCCAT2000* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_CAT02`: *CAT02* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_CAT02_BRILL2008`: *Brill and Susstrunk (2008)*
    corrected CAT02 chromatic adaptation transform.
-   :attr:`colour.adaptation.CAT_CAT16`: *CAT16* chromatic adaptation
    transform.
-   :attr:`colour.adaptation.CAT_BIANCO2010`: *Bianco and Schettini (2010)*
    chromatic adaptation transform.
-   :attr:`colour.adaptation.CAT_PC_BIANCO2010`:
    *Bianco and Schettini PC (2010)* chromatic adaptation transform.

References
----------
-   :cite:`Bianco2010a` : Bianco, S., & Schettini, R. (2010). Two new von Kries
    based chromatic adaptation transforms found by numerical optimization.
    Color Research & Application, 35(3), 184-192. doi:10.1002/col.20573
-   :cite:`Brill2008a` : Brill, M. H., & Susstrunk, S. (2008). Repairing gamut
    problems in CIECAM02: A progress report. Color Research & Application,
    33(5), 424-426. doi:10.1002/col.20432
-   :cite:`CIETC1-321994b` : CIE TC 1-32. (1994). CIE 109-1994 A Method of
    Predicting Corresponding Colours under Different Chromatic and Illuminance
    Adaptations. Commission Internationale de l'Eclairage.
    ISBN:978-3-900734-51-0
-   :cite:`Fairchild2013ba` : Fairchild, M. D. (2013). The Nayatani et al.
    Model. In Color Appearance Models (3rd ed., pp. 4810-5085). Wiley.
    ISBN:B00DAYO8E2
-   :cite:`Fairchildb` : Fairchild, M. D. (n.d.). Fairchild YSh.
    http://rit-mcsl.org/fairchild//files/FairchildYSh.zip
-   :cite:`Li2007e` : Li, C., Perales, E., Luo, M. R., & Martinez-verdu, F.
    (2007). The Problem with CAT02 and Its Correction.
    https://pdfs.semanticscholar.org/b5a9/\
0215ad9a1fb6b01f310b3d64305f7c9feb3a.pdf
-   :cite:`Li2017` : Li, C., Li, Z., Wang, Z., Xu, Y., Luo, M. R., Cui, G.,
    Melgosa, M., Brill, M. H., & Pointer, M. (2017). Comprehensive color
    solutions: CAM16, CAT16, and CAM16-UCS. Color Research & Application,
    42(6), 703-718. doi:10.1002/col.22131
-   :cite:`Lindbloom2009g` : Fairchild, M. D. (2013). Chromatic Adaptation
    Models. In Color Appearance Models (3rd ed., pp. 4179-4252). Wiley.
    ISBN:B00DAYO8E2
-   :cite:`Nayatani1995a` : Nayatani, Y., Sobagaki, H., & Yano, K. H. T.
    (1995). Lightness dependency of chroma scales of a nonlinear
    color-appearance model and its latest formulation. Color Research &
    Application, 20(3), 156-167. doi:10.1002/col.5080200305
-   :cite:`Westland2012g` : Westland, S., Ripamonti, C., & Cheung, V. (2012).
    CMCCAT97. In Computational Colour Science Using MATLAB (2nd ed., p. 80).
    ISBN:978-0-470-66569-5
-   :cite:`Westland2012k` : Westland, S., Ripamonti, C., & Cheung, V. (2012).
    CMCCAT2000. In Computational Colour Science Using MATLAB (2nd ed., pp.
    83-86). ISBN:978-0-470-66569-5
-   :cite:`Wikipedia2007` : Wikipedia. (2007). CAT02. Retrieved February 24,
    2014, from http://en.wikipedia.org/wiki/CIECAM02#CAT02
"""

from __future__ import annotations

import numpy as np

from colour.hints import NDArray
from colour.utilities import CaseInsensitiveMapping

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CAT_XYZ_SCALING",
    "CAT_VON_KRIES",
    "CAT_BRADFORD",
    "CAT_SHARP",
    "CAT_FAIRCHILD",
    "CAT_CMCCAT97",
    "CAT_CMCCAT2000",
    "CAT_CAT02",
    "CAT_CAT02_BRILL2008",
    "CAT_CAT16",
    "CAT_BIANCO2010",
    "CAT_PC_BIANCO2010",
    "CHROMATIC_ADAPTATION_TRANSFORMS",
]

CAT_XYZ_SCALING: NDArray = np.array(np.identity(3)).reshape([3, 3])
"""
*XYZ Scaling* chromatic adaptation transform.

References
----------
:cite:`Lindbloom2009g`
"""

CAT_VON_KRIES: NDArray = np.array(
    [
        [0.4002400, 0.7076000, -0.0808100],
        [-0.2263000, 1.1653200, 0.0457000],
        [0.0000000, 0.0000000, 0.9182200],
    ]
)
"""
*Von Kries* chromatic adaptation transform.

References
----------
:cite:`CIETC1-321994b`, :cite:`Fairchild2013ba`, :cite:`Lindbloom2009g`,
:cite:`Nayatani1995a`
"""

CAT_BRADFORD: NDArray = np.array(
    [
        [0.8951000, 0.2664000, -0.1614000],
        [-0.7502000, 1.7135000, 0.0367000],
        [0.0389000, -0.0685000, 1.0296000],
    ]
)
"""
*Bradford* chromatic adaptation transform.

References
----------
:cite:`Lindbloom2009g`
"""

CAT_SHARP: NDArray = np.array(
    [
        [1.2694, -0.0988, -0.1706],
        [-0.8364, 1.8006, 0.0357],
        [0.0297, -0.0315, 1.0018],
    ]
)
"""
*Sharp* chromatic adaptation transform.

References
----------
:cite:`Bianco2010a`
"""

CAT_FAIRCHILD: NDArray = np.array(
    [
        [0.8562, 0.3372, -0.1934],
        [-0.8360, 1.8327, 0.0033],
        [0.0357, -0.0469, 1.0112],
    ]
)
"""
*Fairchild* chromatic adaptation transform.

References
----------
:cite:`Fairchildb`
"""

CAT_CMCCAT97: NDArray = np.array(
    [
        [0.8951, -0.7502, 0.0389],
        [0.2664, 1.7135, 0.0685],
        [-0.1614, 0.0367, 1.0296],
    ]
)
"""
*CMCCAT97* chromatic adaptation transform.

References
----------
:cite:`Westland2012g`
"""

CAT_CMCCAT2000: NDArray = np.array(
    [
        [0.7982, 0.3389, -0.1371],
        [-0.5918, 1.5512, 0.0406],
        [0.0008, 0.0239, 0.9753],
    ]
)
"""
*CMCCAT2000* chromatic adaptation transform.

References
----------
:cite:`Westland2012k`
"""

CAT_CAT02: NDArray = np.array(
    [
        [0.7328, 0.4296, -0.1624],
        [-0.7036, 1.6975, 0.0061],
        [0.0030, 0.0136, 0.9834],
    ]
)
"""
*CAT02* chromatic adaptation transform.

References
----------
:cite:`Wikipedia2007`
"""

CAT_CAT02_BRILL2008: NDArray = np.array(
    [
        [0.7328, 0.4296, -0.1624],
        [-0.7036, 1.6975, 0.0061],
        [0.0000, 0.0000, 1.0000],
    ]
)
"""
*Brill and Susstrunk (2008)* corrected CAT02 chromatic adaptation
transform.

References
----------
:cite:`Brill2008a`, :cite:`Li2007e`
"""

CAT_CAT16: NDArray = np.array(
    [
        [0.401288, 0.650173, -0.051461],
        [-0.250268, 1.204414, 0.045854],
        [-0.002079, 0.048952, 0.953127],
    ]
)
"""
*CAT16* chromatic adaptation transform.

References
----------
:cite:`Li2017`
"""

CAT_BIANCO2010: NDArray = np.array(
    [
        [0.8752, 0.2787, -0.1539],
        [-0.8904, 1.8709, 0.0195],
        [-0.0061, 0.0162, 0.9899],
    ]
)
"""
*Bianco and Schettini (2010)* chromatic adaptation transform.

References
----------
:cite:`Bianco2010a`
"""

CAT_PC_BIANCO2010: NDArray = np.array(
    [
        [0.6489, 0.3915, -0.0404],
        [-0.3775, 1.3055, 0.0720],
        [-0.0271, 0.0888, 0.9383],
    ]
)
"""
*Bianco and Schettini PC (2010)* chromatic adaptation transform.

References
----------
:cite:`Bianco2010a`

Notes
-----
-   This chromatic adaptation transform has no negative lobes.
"""

CHROMATIC_ADAPTATION_TRANSFORMS: CaseInsensitiveMapping = (
    CaseInsensitiveMapping(
        {
            "XYZ Scaling": CAT_XYZ_SCALING,
            "Von Kries": CAT_VON_KRIES,
            "Bradford": CAT_BRADFORD,
            "Sharp": CAT_SHARP,
            "Fairchild": CAT_FAIRCHILD,
            "CMCCAT97": CAT_CMCCAT97,
            "CMCCAT2000": CAT_CMCCAT2000,
            "CAT02": CAT_CAT02,
            "CAT02 Brill 2008": CAT_CAT02_BRILL2008,
            "CAT16": CAT_CAT16,
            "Bianco 2010": CAT_BIANCO2010,
            "Bianco PC 2010": CAT_PC_BIANCO2010,
        }
    )
)
CHROMATIC_ADAPTATION_TRANSFORMS.__doc__ = """
Chromatic adaptation transforms.

References
----------
:cite:`Bianco2010a`, :cite:`Brill2008a`, :cite:`Fairchildb`, :cite:`Li2007e`,
:cite:`Li2017`, :cite:`Lindbloom2009g`, :cite:`Westland2012g`,
:cite:`Westland2012k`, :cite:`Wikipedia2007`
"""



"""Showcases *Lightness* computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box('"Lightness" Computations')

xyY = np.array([0.4316, 0.3777, 0.1008])
message_box(
    f'Computing "Lightness" "CIE L*a*b*" reference value for given "CIE xyY" '
    f"colourspace values:\n\n\t{xyY}"
)
print(np.ravel(colour.XYZ_to_Lab(colour.xyY_to_XYZ(xyY)))[0])

print("\n")

Y = 12.19722535
message_box(
    f'Computing "Lightness" using '
    f'"Glasser, Mckinney, Reilly and Schnelle (1958)" method for given '
    f'"luminance" value:\n\n\t{Y}'
)
print(colour.lightness(Y, method="Glasser 1958"))
print(colour.colorimetry.lightness_Glasser1958(Y))

print("\n")

message_box(
    f'Computing "Lightness" using "Wyszecki (1963)" method for given '
    f'"luminance" value:\n\n\t{Y}'
)
print(colour.lightness(Y, method="Wyszecki 1963"))
print(colour.colorimetry.lightness_Wyszecki1963(Y))

print("\n")

message_box(
    f'Computing "Lightness" using "Fairchild and Wyble (2010)" method for '
    f'given "luminance" value:\n\n\t{Y}'
)
print(colour.lightness(Y, method="Fairchild 2010"))
print(colour.colorimetry.lightness_Fairchild2010(Y / 100.0))

print("\n")

message_box(
    f'Computing "Lightness" using "Fairchild and Chen (2011)" method for '
    f'given "luminance" value:\n\n\t{Y}'
)
print(colour.lightness(Y, method="Fairchild 2011"))
print(colour.colorimetry.lightness_Fairchild2011(Y / 100.0))

print("\n")

message_box(
    f'Computing "Lightness" using "CIE 1976" method for given "luminance" '
    f"value:\n\n\t{Y}"
)
print(colour.lightness(Y, method="CIE 1976"))
print(colour.colorimetry.lightness_CIE1976(Y))



"""
RIMM, ROMM and ERIMM Encodings
==============================

Defines the *RIMM, ROMM and ERIMM* encodings:

-   :attr:`colour.models.RGB_COLOURSPACE_ROMM_RGB`.
-   :attr:`colour.models.RGB_COLOURSPACE_RIMM_RGB`.
-   :attr:`colour.models.RGB_COLOURSPACE_ERIMM_RGB`.
-   :attr:`colour.models.RGB_COLOURSPACE_PROPHOTO_RGB`.

References
----------
-   :cite:`ANSI2003a` : ANSI. (2003). Specification of ROMM RGB (pp. 1-2).
    http://www.color.org/ROMMRGB.pdf
-   :cite:`Spaulding2000b` : Spaulding, K. E., Woolfe, G. J., & Giorgianni, E.
    J. (2000). Reference Input/Output Medium Metric RGB Color Encodings
    (RIMM/ROMM RGB) (pp. 1-8). http://www.photo-lovers.org/pdf/color/romm.pdf
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    cctf_encoding_ROMMRGB,
    cctf_decoding_ROMMRGB,
    cctf_encoding_RIMMRGB,
    cctf_decoding_RIMMRGB,
    log_encoding_ERIMMRGB,
    log_decoding_ERIMMRGB,
)


__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_ROMM_RGB",
    "WHITEPOINT_NAME_ROMM_RGB",
    "CCS_WHITEPOINT_ROMM_RGB",
    "MATRIX_ROMM_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_ROMM_RGB",
    "RGB_COLOURSPACE_ROMM_RGB",
    "PRIMARIES_RIMM_RGB",
    "WHITEPOINT_NAME_RIMM_RGB",
    "CCS_WHITEPOINT_RIMM_RGB",
    "MATRIX_RIMM_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_RIMM_RGB",
    "RGB_COLOURSPACE_RIMM_RGB",
    "PRIMARIES_ERIMM_RGB",
    "WHITEPOINT_NAME_ERIMM_RGB",
    "CCS_WHITEPOINT_ERIMM_RGB",
    "MATRIX_ERIMM_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_ERIMM_RGB",
    "RGB_COLOURSPACE_ERIMM_RGB",
    "PRIMARIES_PROPHOTO_RGB",
    "WHITEPOINT_NAME_PROPHOTO_RGB",
    "CCS_WHITEPOINT_PROPHOTO_RGB",
    "MATRIX_PROPHOTO_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_PROPHOTO_RGB",
    "RGB_COLOURSPACE_PROPHOTO_RGB",
]

PRIMARIES_ROMM_RGB: NDArray = np.array(
    [
        [0.7347, 0.2653],
        [0.1596, 0.8404],
        [0.0366, 0.0001],
    ]
)
"""*ROMM RGB* colourspace primaries."""

WHITEPOINT_NAME_ROMM_RGB: str = "D50"
"""*ROMM RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_ROMM_RGB: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_ROMM_RGB]
"""*ROMM RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_ROMM_RGB_TO_XYZ: NDArray = np.array(
    [
        [0.7977, 0.1352, 0.0313],
        [0.2880, 0.7119, 0.0001],
        [0.0000, 0.0000, 0.8249],
    ]
)
"""*ROMM RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_ROMM_RGB: NDArray = np.array(
    [
        [1.3460, -0.2556, -0.0511],
        [-0.5446, 1.5082, 0.0205],
        [0.0000, 0.0000, 1.2123],
    ]
)
"""*CIE XYZ* tristimulus values to *ROMM RGB* colourspace matrix."""

RGB_COLOURSPACE_ROMM_RGB: RGB_Colourspace = RGB_Colourspace(
    "ROMM RGB",
    PRIMARIES_ROMM_RGB,
    CCS_WHITEPOINT_ROMM_RGB,
    WHITEPOINT_NAME_ROMM_RGB,
    MATRIX_ROMM_RGB_TO_XYZ,
    MATRIX_XYZ_TO_ROMM_RGB,
    cctf_encoding_ROMMRGB,
    cctf_decoding_ROMMRGB,
)
RGB_COLOURSPACE_ROMM_RGB.__doc__ = """
*ROMM RGB* colourspace.

References
----------
:cite:`ANSI2003a`, :cite:`Spaulding2000b`
"""

PRIMARIES_RIMM_RGB: NDArray = PRIMARIES_ROMM_RGB
"""*RIMM RGB* colourspace primaries."""

WHITEPOINT_NAME_RIMM_RGB: str = WHITEPOINT_NAME_ROMM_RGB
"""*RIMM RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_RIMM_RGB: NDArray = CCS_WHITEPOINT_ROMM_RGB
"""*RIMM RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_RIMM_RGB_TO_XYZ: NDArray = MATRIX_ROMM_RGB_TO_XYZ
"""*RIMM RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_RIMM_RGB: NDArray = MATRIX_XYZ_TO_ROMM_RGB
"""*CIE XYZ* tristimulus values to *RIMM RGB* colourspace matrix."""

RGB_COLOURSPACE_RIMM_RGB: RGB_Colourspace = RGB_Colourspace(
    "RIMM RGB",
    PRIMARIES_RIMM_RGB,
    CCS_WHITEPOINT_RIMM_RGB,
    WHITEPOINT_NAME_RIMM_RGB,
    MATRIX_RIMM_RGB_TO_XYZ,
    MATRIX_XYZ_TO_RIMM_RGB,
    cctf_encoding_RIMMRGB,
    cctf_decoding_RIMMRGB,
)
RGB_COLOURSPACE_RIMM_RGB.__doc__ = """
*RIMM RGB* colourspace. In cases in which it is necessary to identify a
specific precision level, the notation *RIMM8 RGB*, *RIMM12 RGB* and
*RIMM16 RGB* is used.

References
----------
:cite:`Spaulding2000b`
"""

PRIMARIES_ERIMM_RGB: NDArray = PRIMARIES_ROMM_RGB
"""*ERIMM RGB* colourspace primaries."""

WHITEPOINT_NAME_ERIMM_RGB: str = WHITEPOINT_NAME_ROMM_RGB
"""*ERIMM RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_ERIMM_RGB: NDArray = CCS_WHITEPOINT_ROMM_RGB
"""*ERIMM RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_ERIMM_RGB_TO_XYZ: NDArray = MATRIX_ROMM_RGB_TO_XYZ
"""*ERIMM RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_ERIMM_RGB: NDArray = MATRIX_XYZ_TO_ROMM_RGB
"""*CIE XYZ* tristimulus values to *ERIMM RGB* colourspace matrix."""

RGB_COLOURSPACE_ERIMM_RGB: RGB_Colourspace = RGB_Colourspace(
    "ERIMM RGB",
    PRIMARIES_ERIMM_RGB,
    CCS_WHITEPOINT_ERIMM_RGB,
    WHITEPOINT_NAME_ERIMM_RGB,
    MATRIX_ERIMM_RGB_TO_XYZ,
    MATRIX_XYZ_TO_ERIMM_RGB,
    log_encoding_ERIMMRGB,
    log_decoding_ERIMMRGB,
)
RGB_COLOURSPACE_ERIMM_RGB.__doc__ = """
*ERIMM RGB* colourspace.

References
----------
:cite:`Spaulding2000b`
"""

PRIMARIES_PROPHOTO_RGB: NDArray = PRIMARIES_ROMM_RGB
"""*ProPhoto RGB* colourspace primaries."""

WHITEPOINT_NAME_PROPHOTO_RGB: str = WHITEPOINT_NAME_ROMM_RGB
"""*ProPhoto RGB* colourspace whitepoint name."""

CCS_WHITEPOINT_PROPHOTO_RGB: NDArray = CCS_WHITEPOINT_ROMM_RGB
"""*ProPhoto RGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_PROPHOTO_RGB_TO_XYZ: NDArray = MATRIX_ROMM_RGB_TO_XYZ
"""*ProPhoto RGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_PROPHOTO_RGB: NDArray = MATRIX_XYZ_TO_ROMM_RGB
"""*CIE XYZ* tristimulus values to *ProPhoto RGB* colourspace matrix."""

RGB_COLOURSPACE_PROPHOTO_RGB: RGB_Colourspace = RGB_Colourspace(
    "ProPhoto RGB",
    PRIMARIES_PROPHOTO_RGB,
    CCS_WHITEPOINT_PROPHOTO_RGB,
    WHITEPOINT_NAME_PROPHOTO_RGB,
    MATRIX_PROPHOTO_RGB_TO_XYZ,
    MATRIX_XYZ_TO_PROPHOTO_RGB,
    # NOTE: The "colour.models.cctf_encoding_ProPhotoRGB" and
    # "colour.models.cctf_decoding_ProPhotoRGB" definitions cannot be pickled
    # properly within that class.
    cctf_encoding_ROMMRGB,
    cctf_decoding_ROMMRGB,
)
RGB_COLOURSPACE_PROPHOTO_RGB.__doc__ = """
*ProPhoto RGB* colourspace, an alias colourspace for *ROMM RGB*.

References
----------
:cite:`ANSI2003a`, :cite:`Spaulding2000b`
"""



"""
Smits (1999) - Reflectance Recovery
===================================

Defines the objects for reflectance recovery using *Smits (1999)* method.

References
----------
-   :cite:`Smits1999a` : Smits, B. (1999). An RGB-to-Spectrum Conversion for
    Reflectances. Journal of Graphics Tools, 4(4), 11-22.
    doi:10.1080/10867651.1999.10487511
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS, SpectralDistribution
from colour.hints import ArrayLike, NDArray
from colour.models import (
    XYZ_to_RGB,
    normalised_primary_matrix,
    RGB_COLOURSPACE_sRGB,
)
from colour.recovery import SDS_SMITS1999
from colour.utilities import to_domain_1

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_SMITS1999",
    "CCS_WHITEPOINT_SMITS1999",
    "MATRIX_XYZ_TO_RGB_SMITS1999",
    "XYZ_to_RGB_Smits1999",
    "RGB_to_sd_Smits1999",
]

PRIMARIES_SMITS1999: NDArray = RGB_COLOURSPACE_sRGB.primaries
"""Current *Smits (1999)* method implementation colourspace primaries."""

CCS_WHITEPOINT_SMITS1999: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
]["E"]
"""Current *Smits (1999)* method implementation colourspace whitepoint."""

MATRIX_XYZ_TO_RGB_SMITS1999: NDArray = np.linalg.inv(
    normalised_primary_matrix(PRIMARIES_SMITS1999, CCS_WHITEPOINT_SMITS1999)
)
"""
Current *Smits (1999)* method implementation *RGB* colourspace to
*CIE XYZ* tristimulus values matrix.
"""


def XYZ_to_RGB_Smits1999(XYZ: ArrayLike) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *RGB* colourspace with
    conditions required by the current *Smits (1999)* method implementation.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colour array.

    Examples
    --------
    >>> XYZ = np.array([0.21781186, 0.12541048, 0.04697113])
    >>> XYZ_to_RGB_Smits1999(XYZ)  # doctest: +ELLIPSIS
    array([ 0.4063959...,  0.0275289...,  0.0398219...])
    """

    return XYZ_to_RGB(
        XYZ,
        CCS_WHITEPOINT_SMITS1999,
        CCS_WHITEPOINT_SMITS1999,
        MATRIX_XYZ_TO_RGB_SMITS1999,
    )


def RGB_to_sd_Smits1999(RGB: ArrayLike) -> SpectralDistribution:
    """
    Recover the spectral distribution of given *RGB* colourspace array using
    *Smits (1999)* method.

    Parameters
    ----------
    RGB
        *RGB* colourspace array to recover the spectral distribution from.

    Returns
    -------
    :class:`colour.SpectralDistribution`
        Recovered spectral distribution.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Smits1999a`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS, SpectralShape
    >>> from colour.colorimetry import sd_to_XYZ_integration
    >>> from colour.utilities import numpy_print_options
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> RGB = XYZ_to_RGB_Smits1999(XYZ)
    >>> cmfs = (
    ...     MSDS_CMFS['CIE 1931 2 Degree Standard Observer'].
    ...     copy().align(SpectralShape(360, 780, 10))
    ... )
    >>> illuminant = SDS_ILLUMINANTS['E'].copy().align(cmfs.shape)
    >>> sd = RGB_to_sd_Smits1999(RGB)
    >>> with numpy_print_options(suppress=True):
    ...     sd # doctest: +ELLIPSIS
    SpectralDistribution([[ 380.        ,    0.0787830...],
                          [ 417.7778    ,    0.0622018...],
                          [ 455.5556    ,    0.0446206...],
                          [ 493.3333    ,    0.0352220...],
                          [ 531.1111    ,    0.0324149...],
                          [ 568.8889    ,    0.0330105...],
                          [ 606.6667    ,    0.3207115...],
                          [ 644.4444    ,    0.3836164...],
                          [ 682.2222    ,    0.3836164...],
                          [ 720.        ,    0.3835649...]],
                         interpolator=LinearInterpolator,
                         interpolator_kwargs={},
                         extrapolator=Extrapolator,
                         extrapolator_kwargs={...})
    >>> sd_to_XYZ_integration(sd, cmfs, illuminant) / 100  # doctest: +ELLIPSIS
    array([ 0.1894770...,  0.1126470...,  0.0474420...])
    """

    white_sd = SDS_SMITS1999["white"].copy()
    cyan_sd = SDS_SMITS1999["cyan"].copy()
    magenta_sd = SDS_SMITS1999["magenta"].copy()
    yellow_sd = SDS_SMITS1999["yellow"].copy()
    red_sd = SDS_SMITS1999["red"].copy()
    green_sd = SDS_SMITS1999["green"].copy()
    blue_sd = SDS_SMITS1999["blue"].copy()

    R, G, B = to_domain_1(RGB)
    sd = white_sd.copy() * 0
    sd.name = f"Smits (1999) - {RGB!r}"

    if R <= G and R <= B:
        sd += white_sd * R
        if G <= B:
            sd += cyan_sd * (G - R)
            sd += blue_sd * (B - G)
        else:
            sd += cyan_sd * (B - R)
            sd += green_sd * (G - B)
    elif G <= R and G <= B:
        sd += white_sd * G
        if R <= B:
            sd += magenta_sd * (R - G)
            sd += blue_sd * (B - R)
        else:
            sd += magenta_sd * (B - G)
            sd += red_sd * (R - B)
    else:
        sd += white_sd * B
        if R <= G:
            sd += yellow_sd * (R - B)
            sd += green_sd * (G - R)
        else:
            sd += yellow_sd * (G - B)
            sd += red_sd * (R - G)

    return sd
